/*  Gringotts - a small utility to safe-keep sensitive data
 *  (c) 2002, Germano Rizzo <mano@pluto.linux.it>
 *
 *  libgrg_crypt.c - routines for data encryption and writing
 *  Author: Germano Rizzo
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <mcrypt.h>
#include <mhash.h>
#include <zlib.h>
#include <bzlib.h>

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <stddef.h>

#include "libgrg_crypt.h"
#include "libgrg_crypt_v1.h"
#include "libgrg_utils.h"
#include "libgringotts.h"

static void
reinit_random (GRG_TOKEN gt)
{
	if (gt && gt->rnd)
		fclose (gt->rnd);
	
	if (gt->sec_lvl == GRG_SEC_PARANOIA)
		gt->rnd = fopen ("/dev/random", "r");
	else
		gt->rnd = fopen ("/dev/urandom", "r");

	if (!gt->rnd)
	{
		if (gt->sec_lvl >= GRG_SEC_MAX)
			return;
		else
		{
			unsigned int i;
			int rnd;

			srand (time (NULL));
			for (i = 0; i < getpid (); i++)
				rnd = rand ();
		}
	}
}

GRG_TOKEN
grg_initialize (unsigned char *header, unsigned char header_len, grg_crypt_algo crypt_algo,
		grg_hash_algo hash_algo, grg_comp_algo comp_algo,
		grg_comp_ratio comp_lvl, grg_security_lvl sec_lvl)
{
	GRG_TOKEN ret = (GRG_TOKEN) malloc (sizeof(struct grg_token));

	if(!header)
		return NULL;
	ret->header = grg_strdup (header, header_len);
	ret->hlen = header_len;
	ret->crypt_algo = crypt_algo;
	ret->hash_algo = hash_algo;
	ret->comp_algo = comp_algo;
	ret->comp_lvl = comp_lvl;
	ret->sec_lvl = sec_lvl;

	reinit_random (ret);

	return ret;
}

GRG_TOKEN
grg_initialize_defaults (unsigned char *header, unsigned char header_len)
{
	return grg_initialize (header, header_len, GRG_SERPENT, GRG_RIPEMD_160, GRG_ZLIB, GRG_LVL_BEST, GRG_SEC_MAX);
}

void
grg_deinitialize (GRG_TOKEN gt)
{
	if (gt->rnd)
	{
		fclose (gt->rnd);
		gt->rnd = NULL;
	}

	free (gt->header);

	gt->hlen = 0;

	free(gt);
}

grg_crypt_algo
grg_tok_get_crypt_algo (GRG_TOKEN gt)
{
	return gt->crypt_algo;
}

grg_hash_algo
grg_tok_get_hash_algo (GRG_TOKEN gt)
{
	return gt->hash_algo;
}

grg_comp_algo
grg_tok_get_comp_algo (GRG_TOKEN gt)
{
	return gt->comp_algo;
}

grg_comp_ratio
grg_tok_get_comp_ratio (GRG_TOKEN gt)
{
	return gt->comp_lvl;
}

grg_security_lvl
grg_tok_get_security_lvl (GRG_TOKEN gt)
{
	return gt->sec_lvl;
}

void
grg_tok_set_crypt_algo (GRG_TOKEN gt, grg_crypt_algo crypt_algo)
{
	gt->crypt_algo = crypt_algo;
}

void
grg_tok_set_hash_algo (GRG_TOKEN gt, grg_hash_algo hash_algo)
{
	gt->hash_algo = hash_algo;
}

void
grg_tok_set_comp_algo (GRG_TOKEN gt, grg_comp_algo comp_algo)
{
	gt->comp_algo = comp_algo;
}

void
grg_tok_set_comp_ratio (GRG_TOKEN gt, grg_comp_ratio comp_ratio)
{
	gt->comp_lvl = comp_ratio;
}

void
grg_tok_set_security_lvl (GRG_TOKEN gt, grg_security_lvl sec_level)
{
	gt->sec_lvl = sec_level;
	reinit_random (gt);
}

/**
 * get_key:
 * @pwd: the password to generate hash key from
 * @pwd_len: its length, in bytes
 * @key_size: the size of the key to produce
 * @hash_algo: if 0, use SHA-1; if 1, RIPEMD-160
 *
 * Produces a key, for use with libmcrypt's encryption engine.
 *
 * Returns: a string (key_size byte) with the key, to be freed after use
 */
static unsigned char *
get_key (unsigned char *pwd, unsigned long pwdlen, unsigned int key_size,
	 grg_hash_algo hash_algo)
{
	unsigned char *ret;

	ret = calloc (1, key_size);

	mhash_keygen (KEYGEN_MCRYPT,
		      (hash_algo ? MHASH_RIPEMD160 : MHASH_SHA1), 0, ret,
		      key_size, NULL, 0, pwd, pwdlen);

	return ret;
}

/**
 * get_CRC32:
 * @string: a byte sequence
 * @strlen: the length of the string
 *
 * Computes the CRC32 checksum of a byte sequence.
 *
 * Returns: the checksum
 */
static unsigned char *
get_CRC32 (unsigned char *string, unsigned long strlen)
{
	MHASH td;
	unsigned char *ret;

	td = mhash_init (MHASH_CRC32);

	if (td == MHASH_FAILED)
		exit (1);

	mhash (td, string, strlen);

	ret = mhash_end (td);

	return ret;
}

/**
 * compare_CRC32:
 * @CRC: the CRC to compare to
 * @toCheck: the byte sequence to compare the CRC to
 * @len: the byte sequence length
 *
 * Tells if a byte sequence has the provided CRC32, in other words if
 * it's equal to the previously CRC'ed one.
 *
 * Returns: TRUE or FALSE
 */
int
compare_CRC32 (unsigned char *CRC, unsigned char *toCheck, unsigned long len)
{
	unsigned char *CRC2;
	int ret;

	CRC2 = get_CRC32 (toCheck, len);

	ret = !memcmp (CRC, CRC2, GRG_CRC_LEN);

	free (CRC2);

	return ret;
}

static unsigned char *
grg2mcrypt (grg_crypt_algo algo)
{
	switch (algo)
	{
	case GRG_RIJNDAEL_128:
		return MCRYPT_RIJNDAEL_128;

	case GRG_SERPENT:
		return MCRYPT_SERPENT;

	case GRG_TWOFISH:
		return MCRYPT_TWOFISH;

	case GRG_CAST_256:
		return MCRYPT_CAST_256;

	case GRG_SAFERPLUS:
		return MCRYPT_SAFERPLUS;

	case GRG_LOKI97:
		return MCRYPT_LOKI97;

	case GRG_3DES:
		return MCRYPT_3DES;

	case GRG_RIJNDAEL_256:
		return MCRYPT_RIJNDAEL_256;

	default:
		return MCRYPT_SERPENT;
	}
}

/**
 * grg_save_crypted:
 * @origData: a byte sequence to read the data to encrypt and write in the file.
 * @pwd: the password to decode data
 * @fpath: path of the file to read data from
 *
 * Stores data in a Gringotts file
 *
 * Returns: 0 if OK; an error code otherwise (see grg_crypt.h)
 */
int
grg_save_crypted (GRG_TOKEN gt, unsigned char *origData, unsigned long odlen,
		  unsigned char *pwd, unsigned long pwdlen,
		  unsigned char *fpath)
{
	unsigned char *compData, *chunk, *toCRC1, *CRC1, *toEnc, *key, *IV,
		*toCRC2, *CRC2, algo;
	unsigned int dIV, dKey, err, i;
	unsigned long uncDim, compDim;
	FILE *fp;
	MCRYPT mod;

	uncDim = odlen;

	if (gt->comp_lvl)
	{
		if (gt->comp_algo)	//bz2
			compDim =
				(unsigned
				 long) ((((float) uncDim) / 100.0 * 101.0) +
					600.0);
		else		//libz
			compDim =
				(unsigned long) ((((float) uncDim) + 12.0) /
						 100.0 * 101.0);

		compData =
			(unsigned char *) malloc (compDim *
						  sizeof (unsigned char));

		//compress the data
		if (gt->comp_algo)	//bz2
			err = BZ2_bzBuffToBuffCompress ((char *) compData,
							(unsigned int *)
							&compDim, origData,
							uncDim,
							gt->comp_lvl * 3, 0,
							0);
		else
			err = compress2 (compData, &compDim, origData, uncDim,
					 gt->comp_lvl * 3);

		if (err < 0)
			return GRG_WRITE_COMP_ERR;
	}
	else
	{
		compDim = uncDim;
		compData = grg_strdup (origData, uncDim);
	}

	if (compDim > MAX_COMP_SIZE * 1048576l)
		return GRG_WRITE_TOO_BIG_ERR;

	chunk = grg_long2char (uncDim);

	//adds the CRC32 and DATA_LEN field
	toCRC1 = grg_strconcat (chunk, compData, NULL, GRG_DATA_DIM_LEN,
				compDim, 0);

	grg_free (gt, chunk, GRG_DATA_DIM_LEN);
	chunk = NULL;
	grg_free (gt, compData, compDim * sizeof (unsigned char));
	compData = NULL;

	compDim += GRG_DATA_DIM_LEN;

	CRC1 = get_CRC32 (toCRC1, compDim);

	toEnc = grg_strconcat (CRC1, toCRC1, NULL, GRG_CRC_LEN, compDim, 0);

	grg_free (gt, CRC1, GRG_CRC_LEN);
	CRC1 = NULL;
	grg_free (gt, toCRC1, compDim);
	toCRC1 = NULL;

	compDim += GRG_CRC_LEN;

	//encrypts the data
	mod = mcrypt_module_open (grg2mcrypt (gt->crypt_algo), NULL,
				  MCRYPT_CFB, NULL);

	if (mod == MCRYPT_FAILED)
	{
		grg_free (gt, toEnc, compDim);
		toEnc = NULL;
		return GRG_WRITE_ENC_INIT_ERR;
	}

	dKey = mcrypt_enc_get_key_size (mod);
	key = get_key (pwd, pwdlen, dKey, gt->hash_algo);
	dIV = mcrypt_enc_get_iv_size (mod);
	IV = grg_rnd_seq (gt, dIV);
	err = mcrypt_generic_init (mod, key, dKey, IV);
	grg_free (gt, key, dKey);
	key = NULL;

	if (err < 0)
	{
		grg_free (gt, IV, dIV);
		IV = NULL;
		grg_free (gt, toEnc, compDim);
		toEnc = NULL;
		return GRG_WRITE_ENC_INIT_ERR;
	}

	for (i = 0; i < compDim; i++)
		mcrypt_generic (mod, &toEnc[i], 1);

	mcrypt_generic_deinit (mod);
	mcrypt_module_close (mod);

	//adds algorythm and salt 

	algo = (unsigned char) (gt->crypt_algo | gt->hash_algo | gt->
				comp_algo | gt->comp_lvl);

	toCRC2 = grg_strconcat (&algo, IV, toEnc, GRG_ALGO_LEN, dIV, compDim);

	grg_free (gt, IV, dIV);
	IV = NULL;
	grg_free (gt, toEnc, compDim);
	toEnc = NULL;

	compDim += GRG_ALGO_LEN + dIV;

	//calculates the CRC32

	CRC2 = get_CRC32 (toCRC2, compDim);

	//writes it all

	fp = fopen (fpath, "wb");

	if (!fp)
	{
		grg_free (gt, CRC2, GRG_CRC_LEN);
		CRC2 = NULL;
		grg_free (gt, toCRC2, compDim);
		toCRC2 = NULL;
		return GRG_WRITE_FILE_ERR;
	}

	fwrite (gt->header, gt->hlen, 1, fp);
	fwrite (GRG_FILE_VERSION, GRG_FILE_VERSION_LEN, 1, fp);
	fwrite (CRC2, GRG_CRC_LEN, 1, fp);
	fwrite (toCRC2, compDim, 1, fp);

	err = fclose (fp);

	if (err)
	{
		grg_free (gt, CRC2, GRG_CRC_LEN);
		CRC2 = NULL;
		grg_free (gt, toCRC2, compDim);
		toCRC2 = NULL;
		return GRG_WRITE_FILE_ERR;
	}

	sync ();

	//closing
	grg_free (gt, CRC2, GRG_CRC_LEN);
	CRC2 = NULL;
	grg_free (gt, toCRC2, compDim);
	toCRC2 = NULL;

	return GRG_OK;
}

/**
 * grg_load_crypted_v2:
 * @origData: a pointer to a byte sequence to store the data in. It must be freed after use!
 * @pwd: the password to decode data
 * @fpath: path of the file to read data from
 * @just_validate: if 1, it only checks for file validity, without loading it.
 * 
 * Loads the encrypted data stored in a Gringotts file (v.2), or just check for validity
 *
 * Returns: 0 if OK; an error code otherwise (see grg_crypt.h)
 */
static int
grg_load_crypted_v2 (GRG_TOKEN gt, unsigned char **origData,
		     unsigned long *odlen, unsigned char *pwd,
		     unsigned int pwdlen, const unsigned char *fpath,
		     unsigned int just_validate)
{
	unsigned char *CRC32, *CRC32b, *compData, *bcompData, *IV,
		*key, *dimdata;
	unsigned int err, dIV, dKey;
	unsigned long flen, bflen, i;
	FILE *fp;
	MCRYPT mod;

	fp = fopen (fpath, "rb");
	
	if (!fp)
		return GRG_READ_FILE_ERR;
	
	flen = grg_file_size (fp);

	if (flen < GRG_OVERHEAD + GRG_IV_SIZE_MIN + gt->hlen)	//the min total length of header part
		return GRG_READ_FILE_ERR;

	if (flen - (GRG_OVERHEAD + GRG_IV_SIZE_MAX + gt->hlen)> MAX_COMP_SIZE * 1048576l)
		return GRG_READ_TOO_BIG_ERR;

	//skip the already checked part

	fseek (fp, gt->hlen + GRG_FILE_VERSION_LEN, SEEK_SET);

	//checks the 1st CRC

	flen -= gt->hlen + GRG_FILE_VERSION_LEN + GRG_CRC_LEN;

	CRC32 = (unsigned char *) malloc (GRG_CRC_LEN *
					  sizeof (unsigned char));
	compData = (unsigned char *) malloc (flen * sizeof (unsigned char));
	bcompData = compData;
	bflen = flen;

	fread (CRC32, 1, GRG_CRC_LEN, fp);
	fread (compData, 1, flen, fp);

	fclose (fp);

	if (!compare_CRC32 (CRC32, compData, flen))
	{
		grg_free (gt, CRC32, GRG_CRC_LEN);
		CRC32 = NULL;
		grg_free (gt, bcompData, bflen);
		bcompData = NULL;
		return GRG_READ_CRC_ERR;
	}

	grg_free (gt, CRC32, GRG_CRC_LEN);
	CRC32 = NULL;

	if (just_validate == 1)
		return GRG_OK;

	//decodes the algo byte
	gt->crypt_algo = (unsigned char) (compData[0] & GRG_ENCRYPT_MASK);
	gt->hash_algo = (unsigned char) (compData[0] & GRG_HASH_MASK);
	gt->comp_algo = (unsigned char) (compData[0] & GRG_COMP_TYPE_MASK);
	gt->comp_lvl = (unsigned char) (compData[0] & GRG_COMP_LVL_MASK);

	compData += GRG_ALGO_LEN * sizeof (unsigned char);
	flen -= GRG_ALGO_LEN * sizeof (unsigned char);

	//decrypts the encrypted data
	mod = mcrypt_module_open (grg2mcrypt (gt->crypt_algo), NULL,
				  MCRYPT_CFB, NULL);

	if (mod == MCRYPT_FAILED)
	{
		grg_free (gt, bcompData, bflen);
		bcompData = NULL;
		return GRG_READ_ENC_INIT_ERR;
	}

	dKey = mcrypt_enc_get_key_size (mod);
	key = get_key (pwd, pwdlen, dKey, gt->hash_algo);
	dIV = mcrypt_enc_get_iv_size (mod);
	IV = grg_strdup (compData, dIV);

	mcrypt_generic_init (mod, key, dKey, IV);
	grg_free (gt, key, dKey);
	key = NULL;
	grg_free (gt, IV, dIV);
	IV = NULL;
	if (err < 0)
	{
		grg_free (gt, bcompData, bflen);
		bcompData = NULL;
		return GRG_READ_ENC_INIT_ERR;
	}

	compData += dIV * sizeof (unsigned char);
	flen -= dIV * sizeof (unsigned char);

	for (i = 0; i < flen; i++)
		mdecrypt_generic (mod, &compData[i], 1);

	mcrypt_generic_deinit (mod);
	mcrypt_module_close (mod);

	//checks the 2nd CRC32

	CRC32b = grg_strdup (compData, GRG_CRC_LEN);

	compData += GRG_CRC_LEN * sizeof (unsigned char);
	flen -= GRG_CRC_LEN * sizeof (unsigned char);

	if (!compare_CRC32 (CRC32b, compData, flen))
	{
		grg_free (gt, bcompData, bflen);
		bcompData = NULL;
		grg_free (gt, CRC32b, GRG_CRC_LEN);
		CRC32b = NULL;
		return GRG_READ_PWD_ERR;
	}

	grg_free (gt, CRC32b, GRG_CRC_LEN);
	CRC32b = NULL;

	//reads the uncompressed data length

	dimdata = grg_strdup (compData, GRG_DATA_DIM_LEN);

	compData += GRG_DATA_DIM_LEN * sizeof (unsigned char);
	flen -= GRG_DATA_DIM_LEN * sizeof (unsigned char);

	*odlen = grg_char2long (dimdata);

	grg_free (gt, dimdata, GRG_DATA_DIM_LEN);
	dimdata = NULL;

	//uncompress the final data

	if (gt->comp_lvl)
	{
		unsigned char *tmpData =
			(unsigned char *) malloc ((*odlen) *
						  sizeof (unsigned char));

		if (gt->comp_algo)	//bz2
			err = BZ2_bzBuffToBuffDecompress ((char *) tmpData,
							  (unsigned int *)
							  odlen,
							  (char *) compData,
							  flen,
							  USE_BZ2_SMALL_MEM,
							  0);
		else		//zlib
			err = uncompress (tmpData, odlen, compData, flen);

		if (err < 0)
		{
			grg_free (gt, tmpData, *odlen);
			tmpData = NULL;
			grg_free (gt, bcompData, bflen);
			bcompData = NULL;
			return GRG_READ_COMP_ERR;
		}

		*origData = grg_strconcat (tmpData, "", NULL, *odlen, 1, 0);

		grg_free (gt, tmpData, *odlen);
		tmpData = NULL;
	}
	else
		*origData = grg_strconcat (compData, "", NULL, *odlen, 1, 0);

	grg_free (gt, bcompData, bflen);
	bcompData = NULL;

	return GRG_OK;
}

/**
 * grg_load_crypted:
 * @origData: a pointer to a byte sequence to store the data in. It must be freed after use!
 * @pwd: the password to decode data
 * @fpath: path of the file to read data from
 * @just_validate: if 1, it only checks for file validity, without loading it.
 * 
 * Loads the encrypted data stored in a Gringotts file (of any version), or just check for validity
 *
 * Returns: 0 if OK; an error code otherwise (see grg_crypt.h)
 */
int
grg_load_crypted (GRG_TOKEN gt, unsigned char **origData,
		  unsigned long *odlen, unsigned char *pwd,
		  unsigned int pwdlen, const unsigned char *fpath,
		  unsigned int just_validate)
{
	unsigned char vers, *flag;
	FILE *fp;

	fp = fopen (fpath, "rb");

	if (!fp)
		return GRG_READ_FILE_ERR;

	//checks the ID header
	flag = (unsigned char *) malloc (gt->hlen *
					 sizeof (unsigned char));

	fread (flag, 1, gt->hlen, fp);

	if (memcmp (gt->header, flag, gt->hlen))
	{
		grg_free (gt, flag, gt->hlen);
		flag = NULL;
		fclose (fp);
		return GRG_READ_MAGIC_ERR;
	}

	grg_free (gt, flag, gt->hlen);
	flag = NULL;

	//checks the GRG_VERSION

	vers = (unsigned char) fgetc (fp);

	fclose (fp);

	switch (vers)
	{
	case '1':
		return grg_load_crypted_v1 (gt, origData, odlen, pwd, pwdlen,
					    fpath, just_validate);
	case '2':
		return grg_load_crypted_v2 (gt, origData, odlen, pwd, pwdlen,
					    fpath, just_validate);
	default:
		return GRG_READ_UNSUPPORTED_VERSION;
	}
}

unsigned int
grg_get_key_size (GRG_TOKEN gt)
{
	MCRYPT mod;

	mod = mcrypt_module_open (grg2mcrypt (gt->crypt_algo), NULL,
				  MCRYPT_CFB, NULL);

	if (mod == MCRYPT_FAILED)
		return 0;

	return mcrypt_enc_get_key_size (mod);
}

unsigned int
grg_get_block_size (GRG_TOKEN gt)
{
	MCRYPT mod;

	mod = mcrypt_module_open (grg2mcrypt (gt->crypt_algo), NULL,
				  MCRYPT_CFB, NULL);

	if (mod == MCRYPT_FAILED)
		return 0;

	return mcrypt_enc_get_block_size (mod);
}

unsigned int
grg_get_key_size_static (grg_crypt_algo crypt_algo)
{
	MCRYPT mod;

	mod = mcrypt_module_open (grg2mcrypt (crypt_algo), NULL, MCRYPT_CFB,
				  NULL);

	if (mod == MCRYPT_FAILED)
		return 0;

	return mcrypt_enc_get_key_size (mod);
}

unsigned int
grg_get_block_size_static (grg_crypt_algo crypt_algo)
{
	MCRYPT mod;

	mod = mcrypt_module_open (grg2mcrypt (crypt_algo), NULL, MCRYPT_CFB,
				  NULL);

	if (mod == MCRYPT_FAILED)
		return 0;

	return mcrypt_enc_get_block_size (mod);
}
