/*  Gringotts - a small utility to safe-keep sensitive data
 *  (c) 2002, Germano Rizzo <mano@pluto.linux.it>
 *
 *  libgrg_utils.c - utility functions for libgringotts
 *  Author: Germano Rizzo
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "config.h"
#include "libgringotts.h"
#include "libgrg_crypt.h"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include <ctype.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>

// Basically, the following 2 functions are the adaptation of string.h
// functions for byte sequences. Strings are byte sequence that are
// \0-terminated; in this case, this is dangerous, because a \0 can
// occour _inside_ the very sequence. Of course, there's no way to tell
// how long the sequence is, other than "remembering" it, and passing it
// to the functions.

/**
 * grg_strdup
 * @src: the source
 * @len: its length
 *
 * Duplicates a byte sequence into a new one
 *
 * Returns: a newly allocated (to free afterward) byte sequence
 */
inline unsigned char *
grg_strdup (unsigned char *src, unsigned long len)
{
	unsigned char *ret;
	
	if(!src || !len)
		return NULL;

	ret = (unsigned char *) malloc (len * sizeof (unsigned char));
	memcpy (ret, src, len);

	return ret;
}

/**
 * grg_strconcat:
 * @src1: the first byte sequence to concatenate
 * @src2: the second byte sequence to concatenate
 * @src3: the third byte sequence to concatenate
 * @len1: the length of the first sequence
 * @len2: the second of the first sequence
 * @len3: the third of the first sequence
 *
 * Concatenates three byte sequences
 *
 * Returns: a newly allocated (to free afterward) byte sequence
 */
inline unsigned char *
grg_strconcat (unsigned char *src1,
	       unsigned char *src2,
	       unsigned char *src3,
	       unsigned long len1, unsigned long len2, unsigned long len3)
{
	unsigned char *ret, *tmp;

	ret = (unsigned char *) malloc ((len1 + len2 + len3) *
					sizeof (unsigned char));
	tmp = ret;
	memcpy (tmp, src1, len1);
	tmp += len1 * sizeof (unsigned char);
	memcpy (tmp, src2, len2);
	tmp += len2 * sizeof (unsigned char);
	memcpy (tmp, src3, len3);

	return ret;
}

/**
 * grg_long2char:
 * @seed: the long to convert
 *
 * Converts a long into four bytes
 *
 * Returns: a newly allocated 4-bytes sequence, to free after use
 */
unsigned char *
grg_long2char (unsigned long seed)
{
	unsigned char *ret;
	int i;

	ret = (unsigned char *) malloc (4 * sizeof (unsigned char));

	for (i = 3; i >= 0; i--, seed >>= 8)
		ret[i] = seed & 0x0ff;

	return ret;
}

/**
 * grg_char2long:
 * @seed: the 4-char sequence to convert
 *
 * Reverts grg_long2char(), converting back into a long
 *
 * Returns: a long
 */
unsigned long
grg_char2long (unsigned char *seed)
{
	unsigned long ret = 0;
	int i;

	for (i = 3; i >= 0; i--)
		ret |= seed[i] << (3 - i) * 8;

	return ret;
}

/**
 * grg_rnd_seq:
 * @size: the size of the sequence to generate
 *
 * Generates a random sequence of bytes.
 *
 * Returns: a newly-allocated byte sequence
 */
unsigned char *
grg_rnd_seq (GRG_TOKEN gt, unsigned int size)
{
	unsigned char ret[size];

	if (gt->rnd)
	{
		fread (&ret, 1, size, gt->rnd);
		return grg_strdup (ret, size);
	}
	else
	{
		unsigned int i;

		for (i = 0; i < size; i++)
			ret[i] = 1 +
				(unsigned char) (255.0 * rand () /
						 (RAND_MAX + 1.0));
		return grg_strdup (ret, size);
	}
}

/**
 * grg_rnd_chr:
 *
 * Returns a random byte.
 *
 * Returns: a random byte.
 */
unsigned char
grg_rnd_chr (GRG_TOKEN gt)
{
	if (gt->rnd)
		return fgetc (gt->rnd);
	else
		return (unsigned int) (255.0 * rand () / (RAND_MAX + 1.0));
}

/**
 * grg_free:
 * @pntr: pointer to the memory to free
 * @dim: length of the sequence; if -1 it must be NULL-terminated
 *
 * Frees a sequence of bytes, overwriting it with random data
 */
void
grg_free (GRG_TOKEN gt, unsigned char *pntr, long dim)
{
	unsigned int i = 0;

	if (!pntr)
		return;

	if(gt->sec_lvl > GRG_SEC_MIN)
	{
		if (dim < 0)
			while (pntr[i] != 0)
				pntr[i++] = grg_rnd_chr (gt);
		else
		{
			unsigned char *rnd = grg_rnd_seq (gt, dim);
			memcpy (pntr, rnd, dim);
			free (rnd);
		}
	}

	free (pntr);
}

/**
 * grg_file_size:
 * @fpath: the file path
 *
 * Measures the size of a file
 *
 * Returns: a long with the size, -1 if it doesn't exists
 */
unsigned long
grg_file_size (FILE *fd)
{
	struct stat sbuf;

	fstat (fileno(fd), &sbuf);

	return (unsigned long) sbuf.st_size;
}

/**
 * grg_ascii_pwd_quality:
 * @pwd: a non-multibyte string with the password
 * @pwd_len: the maximum length, to grant a termination. If
 * negative, the string _must_ be null-terminated
 *
 * Calculates an indicative value proportional to the password
 * "quality". Returns a value from 0 to 1, where 0 is the worst
 * password (4 char, all numeric), 1 the approximation of 256
 * bits of real data in the password, or better. The scale is 
 * logarythmic, so it will grow faster for shorter passwords than 
 * for larger ones, since improvements in a scarcely-secure password
 * are more valuable than improvements in better ones.<br>
 * Anyway, this should be used only for ASCII-like passwords, null
 * terminated. If you want to eval more "generic" sequences of 
 * bytes, use grg_file_pwd_quality. If you use multibyte encodings,
 * you must convert them to your locale encoding before evaluation.
 * 
 * Returns: a double between 0 and 1, inclusive
 */
double
grg_ascii_pwd_quality (const unsigned char *pwd, long pwd_len)
{
	char A=FALSE, a=FALSE, n=FALSE, p=FALSE;
	long i=0;
	unsigned char basin = 0;
	double ret;

	if(!pwd)
		return 0.0;
	
	if(pwd_len<0)
		pwd_len = LONG_MAX-1;

	while((pwd[i] != '\0') && (i<pwd_len)){

		if(islower(pwd[i])){
			a=TRUE;
			i++;
			continue;
		}

		if(isdigit(pwd[i])){
			n=TRUE;
			i++;
			continue;
		}

		if(isupper(pwd[i])){
			A=TRUE;
			i++;
			continue;
		}
		
		p=TRUE;
		i++;
	}
	
	if(i<4)
		return 0.0;
	
	if(a)
		basin += 26;
	if(A)
		basin += 26;
	if(n)
		basin += 10;
	if(p)
		basin += 32;

	#undef LOG2
	#define LOG2(val) \
		(log(val)/0.693147) 	/* log, basis 2 */
	#define QVAL(length, num_char) \
		(length*LOG2(num_char))	/* number of `real' bits in pwd */
	#define QMAX	5.545177	//log(256)
	#define QMIN	2.586840	//log(4*LOG2(10))
	#define QDIF	2.958337	//QMAX-QMIN
	
	ret = (log(QVAL(i, basin))-QMIN)/QDIF;

	if (ret < 0)
		return 0.0;
	if (ret > 1)
		return 1.0;
	return ret;
}

/**
 * grg_file_pwd_quality:
 * @pwd_path: the path to the file to use as password
 *
 * Calculates an indicative value proportional to the password
 * "quality", in the case you want to use a file content as a
 * password. Returns a value from 0 to 1, where 0 is the worst
 * case (empty file), 1 the approximation of 256 bits of real 
 * data, or better. The scale is linear.
 * 
 * Returns: a double between 0 and 1, inclusive
 */
double
grg_file_pwd_quality (const unsigned char *pwd_path)
{
	double ret;
	FILE *pdf;

	pdf = fopen(pwd_path, "r");

	if (!pdf)
		return 0.0;

	ret = ((double)(grg_file_size (pdf))) / 32.0;

	fclose (pdf);

	if (ret < 0)
		return 0.0;
	if (ret > 1)
		return 1.0;
	return ret;
}

