/*  Gringotts - a small utility to safe-keep sensitive data
 *  (c) 2002, Germano Rizzo <mano@pluto.linux.it>
 *
 *  grg_utils.c - utility functions for gringotts
 *  Author: Germano Rizzo
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "grg_defs.h"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <sys/stat.h>

// Basically, the following 2 functions are the adaptation of string.h
// functions for byte sequences. Strings are byte sequence that are
// \0-terminated; in this case, this is dangerous, because a \0 can
// occour _inside_ the very sequence. Of course, there's no way to tell
// how long the sequence is, other than "remembering" it, and passing it
// to the functions.

/**
 * grg_strdup
 * @src: the source
 * @len: its length
 *
 * Duplicates a byte sequence into a new one
 *
 * Returns: a newly allocated (to free afterward) byte sequence
 */
inline unsigned char *
grg_strdup (unsigned char *src, unsigned long len)
{
	unsigned char *ret;

	ret = (unsigned char *) malloc (len * sizeof (unsigned char));
	memcpy (ret, src, len);

	return ret;
}

/**
 * grg_strconcat:
 * @src1: the first byte sequence to concatenate
 * @src2: the second byte sequence to concatenate
 * @src3: the third byte sequence to concatenate
 * @len1: the length of the first sequence
 * @len2: the second of the first sequence
 * @len3: the third of the first sequence
 *
 * Concatenates three byte sequences
 *
 * Returns: a newly allocated (to free afterward) byte sequence
 */
inline unsigned char *
grg_strconcat (unsigned char *src1,
	       unsigned char *src2,
	       unsigned char *src3,
	       unsigned long len1, unsigned long len2, unsigned long len3)
{
	unsigned char *ret, *tmp;

	ret = (unsigned char *) malloc ((len1 + len2 + len3) *
					sizeof (unsigned char));
	tmp = ret;
	memcpy (tmp, src1, len1);
	tmp += len1 * sizeof (unsigned char);
	memcpy (tmp, src2, len2);
	tmp += len2 * sizeof (unsigned char);
	memcpy (tmp, src3, len3);

	return ret;
}

/**
 * grg_string_endswith:
 * @haystack: the string to search into
 * @needle: the string to search
 *
 * Checks if a NULL-terminated string ands with another one
 *
 * Returns: 0 if false, 1 if true
 */
unsigned int
grg_string_endswith (const unsigned char *haystack, unsigned char *needle)
{
	unsigned int len1 = strlen (haystack), len2 = strlen (needle), i;

	if (len2 > len1)
		return 0;

	for (i = 1; i <= len2; i++)
		if (haystack[len1 - i] != needle[len2 - i])
			return 0;

	return 1;
}

gint
skip_lines (FILE * fp, unsigned int number)
{
	int count = 0, ch;
	while (count < number)
	{
		ch = fgetc (fp);
		if (ch == EOF)
			return FALSE;
		if (ch == '\n')
			count++;
	}
	return TRUE;
}

gchar *
grg_file_rel_to_abs (const gchar *path)
{
	gchar *dir, *file, *cmd, buf[1024];
	FILE *fp = NULL;

	if(g_path_is_absolute(path))
		return g_strdup(path);

	if(!g_file_test (path, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_REGULAR))
		return NULL;

	file = g_path_get_basename (path);
	dir = g_path_get_dirname (path);
	cmd = g_strconcat ("(cd ", dir, "&& echo -e `pwd`\\\\c)", NULL);
	g_free(dir);
	fp = popen(cmd, "r");
	fgets(buf, 1023, fp);
	pclose(fp);
	buf[1023] = '\0';
	g_free(cmd);

	cmd = g_build_filename (buf, file, NULL);
	
	g_free(file);

	return cmd;
}
