/*  Gringotts - a small utility to safe-keep sensitive data
 *  (c) 2002, Germano Rizzo <mano@pluto.linux.it>
 *
 *  gringotts.c - main program and interface
 *  Authors: Germano Rizzo, Nicholas Pouillon
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>

#include <gtk/gtk.h>

#include "../libgringotts/libgringotts.h"

#include "grg_entries.h"
#include "grg_widgets.h"
#include "grg_list.h"
#include "grg_defs.h"
#include "grg_menus.h"
#include "grg_utils.h"
#include "grg_prefs.h"
#include "grg_prefs_io.h"
#include "grg_pwd.h"
#include "grg_safe.h"
#include "grg_recent_dox.h"
#include "grg_popt.h"

/* - some menu buttons are never deactivated (i.e. Quit), so their widgets aren't really needed
 */

// general
static GtkWidget *title, *win1, *edit, *lbl;

// toolbar Navigation
static GtkWidget *bfirst, *bback, *bfor, *blast, *bind;

// others
static GtkWidget *btitle;

// main toolbar 
static GtkWidget *tnew, *topen, *tsave, *tclose, *tquit;
static GtkWidget *tadd, *trem, *tcut, *tcopy, *tpast, *tfind, *tpref;

static guchar *pwd = NULL, *grgfile = NULL, *caption = NULL, *needle = NULL;
static glong pwd_len = 0;
static gboolean started = FALSE, gtk_loop_started = FALSE, created = FALSE;
static GtkClipboard *clip = NULL;
static GtkTextBuffer *entryBuf = NULL;
static gulong entry_ch_sig;

GRG_TOKEN gt = NULL;

/*
 * nyi:
 *
 * Displays a "not yet implemented" message box
 *
void
nyi (void)
{
	grg_msg (_("Sorry, this function hasn't been implemented yet"),
		 GTK_MESSAGE_ERROR, win1);
}*/


/**
 * about:
 *
 * Displays the `About' box.
 */
void
about (void)
{
	gchar *info =
		g_strconcat (GRG_CAP_NAME, _(" version "), GRG_VERSION, "\n",
			     _
			     ("(c) 2002, Germano Rizzo <mano@pluto.linux.it> \n"
			      "Authors:\n"
			      "    Germano Rizzo <mano@pluto.linux.it>\n"
			      "    Nicolas Pouillon <nipo@ssji.net>\n"
			      "\n"
			      "Gringotts is a small but (hopely ;) useful utility that stores "
			      "sensitive data (passwords, credit card numbers, girlfriends' "
			      "addresses) in an organized and most of all very secure form.\n"
			      "It uses libmcrypt and libmhash to provide a strong level of "
			      "encryption, just trying to be as trustworthy as possible.\n\n"
			      "This program is released under the GNU GPL, v.2 or later\n"
			      "See COPYING or go to http://www.gnu.org/copyleft/gpl.html\n\n"
			      "libmcrypt is (c) 1998,1999 Nikos Mavroyanopoulos\n"
			      "libmhash is (c) 2001 Nikos Mavroyanopoulos"), NULL);

	grg_msg (info, GTK_MESSAGE_INFO, win1);
	g_free (info);
}

/**
 * prefs:
 *
 * Displays the preferences window.
 */
void
launch_prefs (void)
{
	grg_pref_dialog (win1);
}

/**
 * quit:
 *
 * Exits the main application cycle, releasing the resources.
 */
static void
quit (void)
{
	grg_free (gt, pwd, pwd_len);
	pwd = NULL;
	g_free (grgfile);
	grgfile = NULL;
	g_free (caption);
	caption = NULL;
	grg_free (gt, needle, -1);
	needle = NULL;
	if (gtk_loop_started)
		gtk_main_quit ();
	grg_prefs_free ();
	grg_security_deinit ();
	grg_recent_dox_deinit ();
	grg_deinitialize (gt);
	exit (0);
}

/**
 * update_saveable:
 * @mode: the operation to perform:
 *        GRG_SAVE_ACTIVE -> activates all the save widgets.
 *        GRG_SAVE_INACTIVE -> deactivates all the save widgets.
 *        GRG_SAVE_QUERY -> query for saveability state.
 *
 * Performs an action about saveability.
 *
 * Returns: if GRG_SAVE_QUERY, the saveability state.
 */
gboolean
update_saveable (grg_saveable mode)
{
	static gboolean saveable = FALSE;

	if (mode == GRG_SAVE_ACTIVE)
	{
		if (started && grgfile)
		{
			caption = g_strconcat (grgfile, " *", NULL);
			gtk_window_set_title (GTK_WINDOW (win1), caption);
		}
		if (!saveable)
		{
			saveable = TRUE;
			gtk_widget_set_sensitive (bsave, TRUE);
			gtk_widget_set_sensitive (tsave, TRUE);
			if (grgfile
			    && (strcmp (grgfile, _("New file")) != 0))
				gtk_widget_set_sensitive (brev, TRUE);
		}
		else
			return saveable;	//no one cares what it returns, actually
	}

	if (mode == GRG_SAVE_INACTIVE)
	{
		if (started && grgfile)
		{
			caption = g_strdup (grgfile);
			gtk_window_set_title (GTK_WINDOW (win1), caption);
		}
		if (saveable == TRUE)
		{
			saveable = FALSE;
			gtk_widget_set_sensitive (bsave, FALSE);
			gtk_widget_set_sensitive (tsave, FALSE);
			gtk_widget_set_sensitive (brev, FALSE);
		}
		else
			return saveable;	//no one cares what it returns, actually
	}
	//let's assume mode is GRG_SAVE_QUERY
	return saveable;
}

/**
 * meta_saveable:
 * @data: unused
 * @user_data: a gpointer to the mode to pass to update_saveable().
 *
 * Controls if update_saveable() can be called, and if yes call it.
 */
static void
meta_saveable (gpointer data, gpointer user_data)
{
	if (started)
		update_saveable ((grg_saveable) GPOINTER_TO_INT (user_data));
}

/**
 * sync_entry:
 *
 * Writes the information stored in the GtkTextBuffer to the
 * entry structure in memory. To be called everytime that a page is
 * leaved, or saved.
 */
static void
sync_entry (void)
{
	static GtkTextIter s, e;

	gtk_text_buffer_get_bounds (entryBuf, &s, &e);
	grg_entries_set_Body (gtk_text_buffer_get_text (entryBuf, &s, &e, FALSE));
}

/*
 * sync_back_entry:
 *
 * Writes the infos in memory to the GtkTextBuffer. To be called whenever
 * a pages is entered from outside gringotts.c.
 *
static void
sync_back_entry (void)
{
	g_signal_handler_block (entryBuf, entry_ch_sig);
	gtk_text_buffer_set_text (entryBuf, grg_entries_get_Body (), -1);
	g_signal_handler_unblock (entryBuf, entry_ch_sig);
}
 */

/**
 * update:
 *
 * Updates various widgets (title of the window, button sensitivity...)
 */
static void
update (void)
{
	static gchar *tlbl;

	if (grg_entries_is_empty ())
	{
		gtk_widget_set_sensitive (btitle, FALSE);
		gtk_widget_set_sensitive (bind, FALSE);
		gtk_widget_set_sensitive (bsas, FALSE);
		gtk_widget_set_sensitive (badd, FALSE);
		gtk_widget_set_sensitive (tadd, FALSE);
		gtk_widget_set_sensitive (bmind, FALSE);
		gtk_widget_set_sensitive (bpwd, FALSE);
		gtk_widget_set_sensitive (tclose, FALSE);
		gtk_widget_set_sensitive (bclose, FALSE);
		gtk_widget_set_sensitive (edit, FALSE);
		gtk_widget_set_sensitive (bcut, FALSE);
		gtk_widget_set_sensitive (tcut, FALSE);
		gtk_widget_set_sensitive (bcop, FALSE);
		gtk_widget_set_sensitive (tcopy, FALSE);
		gtk_widget_set_sensitive (bpaste, FALSE);
		gtk_widget_set_sensitive (tpast, FALSE);
		gtk_widget_set_sensitive (bfind, FALSE);
		gtk_widget_set_sensitive (tfind, FALSE);
		gtk_widget_set_sensitive (bpwd, FALSE);
		gtk_label_set_text (GTK_LABEL (title),
				    GRG_CAP_NAME " " GRG_VERSION);
		g_signal_handler_block (entryBuf, entry_ch_sig);
		gtk_text_buffer_set_text (entryBuf, "", 0);
		g_signal_handler_unblock (entryBuf, entry_ch_sig);		
	}
	else
	{
		gtk_widget_set_sensitive (btitle, TRUE);
		gtk_widget_set_sensitive (bind, TRUE);
		gtk_widget_set_sensitive (bmind, TRUE);
		gtk_widget_set_sensitive (badd, TRUE);
		gtk_widget_set_sensitive (tadd, TRUE);
		gtk_widget_set_sensitive (bsas, TRUE);
		gtk_widget_set_sensitive (bpwd, TRUE);
		gtk_widget_set_sensitive (bclose, TRUE);
		gtk_widget_set_sensitive (tclose, TRUE);
		gtk_widget_set_sensitive (edit, TRUE);
		gtk_widget_set_sensitive (bcut, TRUE);
		gtk_widget_set_sensitive (tcut, TRUE);
		gtk_widget_set_sensitive (bcop, TRUE);
		gtk_widget_set_sensitive (tcopy, TRUE);
		gtk_widget_set_sensitive (bpaste, TRUE);
		gtk_widget_set_sensitive (tpast, TRUE);
		gtk_widget_set_sensitive (bfind, TRUE);
		gtk_widget_set_sensitive (tfind, TRUE);
		gtk_widget_set_sensitive (bpwd, TRUE);
		gtk_label_set_text (GTK_LABEL (title), grg_entries_get_ID ());
		g_signal_handler_block (entryBuf, entry_ch_sig);
		gtk_text_buffer_set_text (entryBuf, grg_entries_get_Body (), -1);
		g_signal_handler_unblock (entryBuf, entry_ch_sig);
	}

	if (grg_entries_is_last ())
	{
		gtk_widget_set_sensitive (bfor, FALSE);
		gtk_widget_set_sensitive (blast, FALSE);
		gtk_widget_set_sensitive (bmfor, FALSE);
		gtk_widget_set_sensitive (bmlast, FALSE);
	}
	else
	{
		gtk_widget_set_sensitive (bfor, TRUE);
		gtk_widget_set_sensitive (blast, TRUE);
		gtk_widget_set_sensitive (bmfor, TRUE);
		gtk_widget_set_sensitive (bmlast, TRUE);
	}

	if (grg_entries_is_first ())
	{
		gtk_widget_set_sensitive (bback, FALSE);
		gtk_widget_set_sensitive (bfirst, FALSE);
		gtk_widget_set_sensitive (bmback, FALSE);
		gtk_widget_set_sensitive (bmfirst, FALSE);
	}
	else
	{
		gtk_widget_set_sensitive (bback, TRUE);
		gtk_widget_set_sensitive (bfirst, TRUE);
		gtk_widget_set_sensitive (bmback, TRUE);
		gtk_widget_set_sensitive (bmfirst, TRUE);
	}

	if (grg_entries_n_el () <= 1)
	{
		gtk_widget_set_sensitive (brem, FALSE);
		gtk_widget_set_sensitive (trem, FALSE);
	}
	else
	{
		gtk_widget_set_sensitive (brem, TRUE);
		gtk_widget_set_sensitive (trem, TRUE);
	}

	if (needle)
		gtk_widget_set_sensitive (bfinda, TRUE);
	else
		gtk_widget_set_sensitive (bfinda, FALSE);

	tlbl = g_strdup_printf ("%d/%d", grg_entries_position () + 1,
				grg_entries_n_el ());
	gtk_label_set_text (GTK_LABEL (lbl), tlbl);
	gtk_window_set_title (GTK_WINDOW (win1), caption);
	g_free (tlbl);
}

void save (void);

/**
 * file_close:
 *
 * Closes the opened document.
 *
 * Returns: GRG_YES if all has gone well, GRG_NO if error, GRG_CANCEL if user canceled
 */
grg_response
file_close (void)
{
	if (grg_entries_is_empty ())
		return GRG_NO;

	if (update_saveable (GRG_SAVE_QUERY))
	{
		grg_response resp = grg_ask_dialog (_("Save?"),
						    _
						    ("Some changes have not been saved.\nDo you wish to save them now?"),
						    TRUE,
						    win1);

		switch (resp)
		{
		case GRG_CANCEL:
			return resp;
		case GRG_YES:
			save ();
		case GRG_NO:
		default:
			break;
		}
	}
	grg_entries_free ();
	grg_free (gt, needle, -1);
	needle = NULL;
	grg_free (gt, pwd, pwd_len);
	pwd = NULL;
	pwd_len = 0;
	g_free (grgfile);
	grgfile = NULL;
	g_free (caption);
	caption = g_strconcat (GRG_CAP_NAME, " ", GRG_VERSION, NULL);
	update_saveable (GRG_SAVE_INACTIVE);
	update ();
	
	return GRG_YES;
}

/**
 * meta_quit:
 *
 * Asks for save if not saved, and then exits.
 */
void
meta_quit (void)
{
	if(file_close () == GRG_CANCEL)
		return;

	quit ();
}

/**
 * revert:
 *
 * Reverts to last saved version.
 */
void
revert (void)
{
	gint err;

	if (update_saveable (GRG_SAVE_QUERY) &&
	    (grg_ask_dialog
	    (_("Confirm"),
	     _("You'll lose all the changes from\nlast save! Are you sure?"),
	     FALSE, win1) == GRG_YES))
	{
		err = grg_entries_load (grgfile, pwd, pwd_len);

		switch (err)
		{
		case GRG_OK:
		{
			update_saveable (GRG_SAVE_INACTIVE);
			update ();
			break;
		}

		case GRG_READ_MAGIC_ERR:
		case GRG_READ_UNSUPPORTED_VERSION:
		case GRG_READ_CRC_ERR:
		case GRG_READ_PWD_ERR:
		case GRG_READ_COMP_ERR:
		case GRG_READ_INVALID_CHARSET_ERR:
		{
			grg_msg (_
				 ("The file appears to be corrupted! Oh, God!"),
				 GTK_MESSAGE_ERROR, win1);
			break;
		}

		case GRG_READ_TOO_BIG_ERR:
		{
			grg_msg (_
				 ("File is too big"),
				 GTK_MESSAGE_ERROR, win1);
			break;
		}

		case GRG_READ_FILE_ERR:
		{
			grg_msg (_("Uh-oh! I can't read from the file!"),
				 GTK_MESSAGE_ERROR, win1);
			break;
		}

		case GRG_READ_ENC_INIT_ERR:
		{
			grg_msg (_
				 ("Problem with libmcrypt, probably a faulty installation"),
				 GTK_MESSAGE_ERROR, win1);
			break;
		}

		default:
			g_assert_not_reached ();
			break;
		}
	}
}

/**
 * load_file:
 * @filename: a filename.
 *
 * Loads a gringotts file.
 */
static void
load_file (gchar * filename)
{
	gchar *tmppwd, *tmpfname;
	gint err, tmppwd_len;
	guchar *res;
	gboolean exit;
	struct stat buf1, buf2;
		
	if (!filename || !*filename)
		return;

	//if this and the opened one are the very same file, fall back on revert()
	if (grgfile &&
	    (strcmp (grgfile, _("New file")) != 0) &&
	    (stat (grgfile, &buf1) == 0) &&
	    (stat (filename, &buf2) == 0) &&
	    (buf1.st_dev == buf2.st_dev) && (buf1.st_ino == buf2.st_ino))
	{
		revert ();
		return;
	}

	if (!g_file_test (filename, G_FILE_TEST_EXISTS))
	{
		grg_msg (_("The selected file doesn't exists"),
			 GTK_MESSAGE_ERROR, win1);
		return;
	}

	if (g_file_test
	    (filename, G_FILE_TEST_IS_DIR | G_FILE_TEST_IS_SYMLINK))
	{
		grg_msg (_("You've selected a directory or a symlink"),
			 GTK_MESSAGE_ERROR, win1);
		return;
	}
	
	tmpfname = grg_file_rel_to_abs (filename);
	g_free(filename);
	filename = g_strdup(tmpfname);
	g_free(tmpfname);

	if(file_close () == GRG_CANCEL)
		return;

	err = grg_load_wrapper (NULL, NULL, 0, filename, 1);

	switch (err)
	{
	case GRG_OK:
		break;

	case GRG_READ_MAGIC_ERR:
	case GRG_READ_UNSUPPORTED_VERSION:
	{
		grg_msg (_
			 ("This file doesn't seem to be a valid Gringotts one!"),
			 GTK_MESSAGE_ERROR, win1);
		return;
	}

	case GRG_READ_FILE_ERR:
	{
		grg_msg (_("Uh-oh! I can't read from the file!"),
			 GTK_MESSAGE_ERROR, win1);
		return;
	}

	case GRG_READ_CRC_ERR:
	case GRG_READ_COMP_ERR:
	{
		grg_msg (_("The file appears to be corrupted! Oh, God!"),
			 GTK_MESSAGE_ERROR, win1);
		return;
	}

	case GRG_READ_TOO_BIG_ERR:
	{
		grg_msg (_
			 ("File is too big"),
			 GTK_MESSAGE_ERROR, win1);
		return;
	}

	default:
	{
		if (err < 0)
		{
			grg_msg (_
				 ("The file appears to be corrupted! Oh, God!"),
				 GTK_MESSAGE_ERROR, win1);
			return;
		}
	}
	}

	while (TRUE)
	{
		exit = FALSE;

		tmppwd = grg_ask_pwd_dialog (win1, &tmppwd_len);

		if (!tmppwd)
			return;
		
#ifdef PWD_HIDING
		//encrypts the password on-the-fly
		grg_security_seal (tmppwd, tmppwd_len);
#endif

		err = grg_load_wrapper (&res, tmppwd, tmppwd_len, filename,
					0);

		switch (err)
		{
		case GRG_OK:
		{
			grg_entries_load_from_string (res);
			grg_free (gt, res, -1);
			res = NULL;
			exit = TRUE;
		}
			break;

		case GRG_READ_PWD_ERR:
		{
			grg_msg (_("Wrong password! Re-enter it"),
				 GTK_MESSAGE_ERROR, win1);
		}
			break;

		case GRG_READ_ENC_INIT_ERR:
		{
			grg_msg (_
				 ("Problem with libmcrypt, probably a faulty installation"),
				 GTK_MESSAGE_ERROR, win1);
		}
			break;

		case GRG_READ_INVALID_CHARSET_ERR:
		{
			grg_msg (_
				 ("Saved data contain invalid UTF-8 chars"),
				 GTK_MESSAGE_ERROR, win1);
		}
			break;

			//just to be sure...
		default:
		{
			if (err < 0)
			{
				grg_msg (_
					 ("The file appears to be corrupted! Oh, God!"),
					 GTK_MESSAGE_ERROR, win1);
				grg_free (gt, tmppwd, tmppwd_len);
				return;
			}
		}
		}
		if (exit)
			break;
	}

	g_free (grgfile);
	grgfile = NULL;

	grgfile = g_strdup (filename);
	grg_free (gt, pwd, pwd_len);
	pwd = g_strdup (tmppwd);
	grg_free (gt, tmppwd, tmppwd_len);
	tmppwd = NULL;
	pwd_len = tmppwd_len;

	g_free (caption);
	caption = g_strdup (grgfile);
	update ();
	update_saveable (GRG_SAVE_INACTIVE);
}

/**
 * meta_load:
 *
 * Displays a file loading dialog, then calls load_file().
 */
void
meta_load (void)
{
	GtkWidget *file_selector;
	gint response;
	file_selector = gtk_file_selection_new (_("Open..."));
	gtk_window_set_transient_for (GTK_WINDOW (file_selector),
				      GTK_WINDOW (win1));
	gtk_widget_show (file_selector);
	response = gtk_dialog_run (GTK_DIALOG (file_selector));
	if (response == GTK_RESPONSE_OK)
	{
		//it may appear a stupid duplication, but it's very important
		//to avoid two file selectors at the same time, so one should
		//be destroyed *before* calling load_file
		gchar *fname = g_strdup (gtk_file_selection_get_filename
					 (GTK_FILE_SELECTION
					  (file_selector)));
		gtk_widget_destroy (file_selector);
		load_file (fname);
		g_free (fname);
	}
	else
		gtk_widget_destroy (file_selector);
}

/**
 * meta_load_file:
 *
 * Calls load_file() with the specified file arg.
 */
void
meta_load_file (gpointer callback_data, guchar * callback_action)
{
	guchar *file = g_strdup (callback_action);

	load_file (file);
	g_free (file);
}

/**
 * move_f:
 *
 * Displays the next entry.
 */
void
move_f (void)
{
	sync_entry ();
	grg_entries_next ();
	update ();
}

/**
 * move_b:
 *
 * Displays the previous entry.
 */
void
move_b (void)
{
	sync_entry ();
	grg_entries_prev ();
	update ();
}

/**
 * move_l:
 *
 * Displays the last entry.
 */
void
move_l (void)
{
	sync_entry ();
	grg_entries_last ();
	update ();
}

/**
 * move_1:
 *
 * Displays the first entry.
 */
void
move_1 (void)
{
	sync_entry ();
	grg_entries_first ();
	update ();
}

/**
 * do_new:
 *
 * Creates a new document.
 */
void
do_new (void)
{
	gchar *tmppwd;
	gint tmppwd_len;

	if(file_close () == GRG_CANCEL)
		return;

	tmppwd = grg_new_pwd_dialog (win1, &tmppwd_len);

	if (!tmppwd)
	{
		grg_msg (_("You must enter a valid password!"),
			 GTK_MESSAGE_ERROR, win1);
		return;
	}
	
#ifdef PWD_HIDING
	//encrypts the password on-the-fly
	grg_security_seal (tmppwd, tmppwd_len);
#endif

	grg_free (gt, pwd, pwd_len);

	g_free (grgfile);
	grgfile = g_strdup (_("New file"));

	if (grg_entries_is_empty ())
		grg_entries_free ();

	pwd = g_strdup (tmppwd);
	grg_free (gt, tmppwd, tmppwd_len);
	tmppwd = NULL;

	pwd_len = tmppwd_len;

	grg_entries_append (_("My first page"));
	grg_entries_first ();
	update_saveable (GRG_SAVE_ACTIVE);
	update ();
}

/**
 * retitle:
 *
 * Gives the entry another title.
 */
static void
retitle (void)
{
	gchar *new = grg_input_dialog (_("New title..."),
				_("New title for this entry:"),
				grg_entries_get_ID (), FALSE, win1);
	if (new)
	{
		grg_entries_set_ID (new);
		grg_free (gt, new, -1);
		new = NULL;
		update_saveable (GRG_SAVE_ACTIVE);
		update ();
	}
}

/**
 * save_as:
 * @fpath: the path to save to
 *
 * Saves a file with another name.
 */
static void
save_as (const gchar * fpath)
{
	gchar *tmpfile;
	gint err;

	tmpfile = g_strdup (fpath);

	if (!grg_string_endswith (fpath, ".grg"))
	{
		guchar *tmp;

		tmp = g_strconcat (tmpfile, ".grg", NULL);
		g_free (tmpfile);
		tmpfile = g_strdup (tmp);
		g_free (tmp);
	}

	if (g_file_test (tmpfile, G_FILE_TEST_EXISTS)
	    && (grg_ask_dialog (_("Overwrite?"),
				_
				("Do you want to overwrite the existing file?"),
				FALSE, win1) != GRG_YES))
	{
		g_free (tmpfile);
		return;
	}

	if (g_file_test (tmpfile, G_FILE_TEST_EXISTS) &&
	    !g_file_test (tmpfile, G_FILE_TEST_IS_REGULAR))
	{
		grg_msg (_("I can't overwrite a directory or a symlink"),
			 GTK_MESSAGE_ERROR, win1);
		g_free (tmpfile);
		return;
	}

	sync_entry ();
	err = grg_entries_save (tmpfile, pwd, pwd_len);

	switch (err)
	{
	case GRG_OK:
	{
		if (grgfile)
			g_free (grgfile);
		grgfile = g_strdup (tmpfile);
		g_free (tmpfile);
		update_saveable (GRG_SAVE_INACTIVE);
		break;
	}

	case GRG_WRITE_COMP_ERR:
	{
		grg_msg (_("Error in compression! Probably a zlib problem"),
			 GTK_MESSAGE_ERROR, win1);
		g_free (tmpfile);
		return;
	}

	case GRG_WRITE_FILE_ERR:
	{
		grg_msg (_("Uh-oh! I can't write to the file!"),
			 GTK_MESSAGE_ERROR, win1);
		g_free (tmpfile);
		return;
	}

	case GRG_WRITE_ENC_INIT_ERR:
	{
		grg_msg (_
			 ("Problem with libmcrypt, probably a faulty installation"),
			 GTK_MESSAGE_ERROR, win1);
		g_free (tmpfile);
		return;
	}

	case GRG_WRITE_TOO_BIG_ERR:
	{
		grg_msg (_
			 ("Too many data to write"), GTK_MESSAGE_ERROR, win1);
		g_free (tmpfile);
		return;
	}

	default:
		g_assert_not_reached ();
		break;
	}
}

/**
 * meta_save_as:
 *
 * Displays the Save As dialog, then calls save_as().
 */
void
meta_save_as (void)
{
	GtkWidget *file_selector;
	gint response;
	file_selector = gtk_file_selection_new (_("Save as...."));
	gtk_window_set_transient_for (GTK_WINDOW (file_selector),
				      GTK_WINDOW (win1));
	gtk_widget_show (file_selector);
	response = gtk_dialog_run (GTK_DIALOG (file_selector));
	if (response == GTK_RESPONSE_OK)
		save_as (gtk_file_selection_get_filename
			 (GTK_FILE_SELECTION (file_selector)));
	gtk_widget_destroy (file_selector);
}

/**
 * save:
 *
 * Saves the current version of the opened document.
 */
void
save (void)
{
	gint err;

	if ((!grgfile) || (strcmp (grgfile, _("New file")) == 0))
	{
		meta_save_as ();
		return;
	}

	sync_entry ();
	err = grg_entries_save (grgfile, pwd, pwd_len);

	switch (err)
	{
	case GRG_OK:
		break;

	case GRG_WRITE_COMP_ERR:
	{
		grg_msg (_("Error in compression! Probably a zlib problem"),
			 GTK_MESSAGE_ERROR, win1);
		return;
	}

	case GRG_WRITE_FILE_ERR:
	{
		grg_msg (_("Uh-oh! I can't write to the file!"),
			 GTK_MESSAGE_ERROR, win1);
		return;
	}

	case GRG_WRITE_ENC_INIT_ERR:
	{
		grg_msg (_
			 ("Problem with libmcrypt, probably a faulty installation"),
			 GTK_MESSAGE_ERROR, win1);
		return;
	}

	case GRG_WRITE_TOO_BIG_ERR:
	{
		grg_msg (_
			 ("Too many data to write"), GTK_MESSAGE_ERROR, win1);
		return;
	}

	default:
		g_assert_not_reached ();
		break;
	}

	update_saveable (GRG_SAVE_INACTIVE);
}

/**
 * insert:
 *
 * Appends a new entry.
 */
void
insert (void)
{
	gchar *new = grg_input_dialog (_("Title..."), _("Title for the new entry:"), _("New"), FALSE, win1);

	sync_entry ();

	if (new)
	{
		grg_entries_append (new);
		grg_free (gt, new, -1);
		new = NULL;
		update_saveable (GRG_SAVE_ACTIVE);
		update ();
	}
}

/**
 * del:
 *
 * Deletes the current entry.
 */
void
del (void)
{
	if (grg_ask_dialog
	    (_("Confirm..."),
	     _("Are you sure you want to remove this entry?"), FALSE, win1) == GRG_YES)
	{
		grg_entries_remove ();
		update_saveable (GRG_SAVE_ACTIVE);
		update ();
	}
}

/**
 * chpwd:
 *
 * Calls the change password dialog.
 */
void
chpwd (void)
{
	gchar *tmppwd, *verpwd;
	gint tmppwd_len, verpwd_len;

	verpwd = grg_ask_pwd_dialog (win1, &verpwd_len);
#ifdef PWD_HIDING
	//encrypts the password on-the-fly
	grg_security_seal (verpwd, verpwd_len);
#endif
	if (!verpwd || (pwd_len != verpwd_len)
	    || (memcmp (pwd, verpwd, pwd_len) != 0))
	{
		grg_msg (_("Wrong password"), GTK_MESSAGE_ERROR, win1);
		grg_free (gt, verpwd, verpwd_len);
		verpwd = NULL;
		return;
	}
	
	grg_free (gt, verpwd, verpwd_len);
	verpwd = NULL;

	tmppwd = grg_new_pwd_dialog (win1, &tmppwd_len);
	
	if (tmppwd)
	{
		grg_free (gt, pwd, pwd_len);
		pwd = g_strdup (tmppwd);
		grg_free (gt, tmppwd, tmppwd_len);
		tmppwd = NULL;
		pwd_len = tmppwd_len;
		tmppwd_len = 0;
#ifdef PWD_HIDING
		//encrypts the password on-the-fly
		grg_security_seal (pwd, pwd_len);
#endif
		grg_msg (_("Password successfully changed"), GTK_MESSAGE_INFO,
			 win1);
		update_saveable (GRG_SAVE_ACTIVE);
	}
}

/**
 * meta_list:
 *
 * Calls the list window.
 */
void
meta_list (void)
{
	sync_entry ();
	grg_list_run ();
	update ();
}

/**
 * cucopa:
 * @callback_data: unused callback param
 * @callback_action: action to perform (GRG_CUT, GRG_COPY, GRG_PASTE)
 *
 * basic CUT/COPY/PASTE clipboard operation.
 */
void
cucopa (gpointer callback_data, guint callback_action)
{
	switch ((grg_clip_action) callback_action)
	{
	case GRG_CUT:
		gtk_text_buffer_cut_clipboard (entryBuf, clip, TRUE);
		return;

	case GRG_COPY:
		gtk_text_buffer_copy_clipboard (entryBuf, clip);
		return;

	case GRG_PASTE:
		gtk_text_buffer_paste_clipboard (entryBuf, clip, NULL, TRUE);
		return;

	default:
		g_assert_not_reached ();
		break;
	}
}

/**
 * find:
 * @callback_data: unused callback param
 * @again: TRUE if I have to continue a previous search
 *
 * Search operation.
 */
void
find (gpointer callback_data, guint again)
{
	static gboolean only_current, case_sens;
	gint offset, found;
	guchar *buf;
	GtkTextIter position;
	GtkTextMark *cursor, *endsel;

	if (!again)
		if (!grg_find_dialog
		    (&needle, &only_current, &case_sens, GTK_WINDOW (win1)))
			return;

	buf = grg_entries_get_Body ();
	//to avoid that searching again and again the same text
	//will find the same portion
	cursor = gtk_text_buffer_get_mark (entryBuf, "selection_bound");
	gtk_text_buffer_get_iter_at_mark (entryBuf, &position, cursor);
	offset = gtk_text_iter_get_offset (&position);

	do
	{
		found = grg_entries_find (needle, offset, only_current,
					  case_sens);

		if (found >= 0)
		{
			buf = grg_entries_get_Body ();
			cursor = gtk_text_buffer_get_mark (entryBuf, "insert");
			gtk_text_buffer_get_iter_at_mark (entryBuf, &position,
							  cursor);
			endsel = gtk_text_buffer_get_mark (entryBuf,
							   "selection_bound");
			gtk_text_iter_set_offset (&position, found);
			gtk_text_buffer_move_mark (entryBuf, cursor, &position);
			gtk_text_iter_set_offset (&position,
						  found +
						  g_utf8_strlen (needle, -1));
			gtk_text_buffer_move_mark (entryBuf, endsel, &position);
			update ();
			break;
		}
		else
		{
			if (only_current)
			{
				grg_msg (_
					 ("The text searched could not be found!"),
					 GTK_MESSAGE_ERROR, win1);
				break;
			}
			else
			{
				if (grg_ask_dialog
				    (_("Wrap around?"),
				     _
				     ("Text not found. Continue search from beginning?"),
				     FALSE, win1) == GRG_YES)
				{
					grg_entries_first ();
					continue;
				}
				else
					break;
			}
		}
	}
	while (TRUE);
}

/**
 * grg_interface:
 *
 * Builds and "activates" the Gringotts main interface.
 */
static void
grg_interface (void)
{
	GtkWidget *vbox, *hbox, *scroll, *menu;
	GtkWidget *tbar_nav, *handle_nav, *tbar_main, *handle_main;
	PangoFontDescription *pfd;

	// window
	win1 = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	grg_window_set_icon (GTK_WINDOW (win1));

	caption = g_strconcat (GRG_CAP_NAME, " ", GRG_VERSION, NULL);

	gtk_container_set_border_width (GTK_CONTAINER (win1), GRG_PAD);
	g_signal_connect (G_OBJECT (win1), "destroy-event",
			  G_CALLBACK (meta_quit), NULL);
	g_signal_connect (G_OBJECT (win1), "delete-event",
			  G_CALLBACK (meta_quit), NULL);
	g_signal_connect (G_OBJECT (win1), "destroy", G_CALLBACK (meta_quit),
			  NULL);
	// the multi-line text widget
	scroll = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scroll),
					     GTK_SHADOW_ETCHED_IN);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scroll),
					GTK_POLICY_AUTOMATIC,
					GTK_POLICY_AUTOMATIC);
	entryBuf = gtk_text_buffer_new (NULL);
	edit = gtk_text_view_new_with_buffer (entryBuf);
	entry_ch_sig = g_signal_connect (G_OBJECT (entryBuf), "changed",
                          G_CALLBACK (meta_saveable),
                          GINT_TO_POINTER (GRG_SAVE_ACTIVE));
	gtk_widget_set_size_request (scroll, 400, 300);
	gtk_container_add (GTK_CONTAINER (scroll), edit);
	gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (edit), GTK_WRAP_WORD);

	// the title widget
	title = gtk_label_new ("");
	gtk_misc_set_alignment (GTK_MISC (title), 0.1, 0.5);
	btitle = gtk_button_new_with_mnemonic (_("_Edit..."));
	g_signal_connect (G_OBJECT (btitle), "clicked", G_CALLBACK (retitle),
			  NULL);
	pfd = pango_font_description_new ();
	pango_font_description_set_weight (pfd, PANGO_WEIGHT_BOLD);
	gtk_widget_modify_font (title, pfd);
	pango_font_description_free (pfd);
	lbl = gtk_label_new ("");

	// the "navigation" lateral toolbar
	tbar_nav = gtk_toolbar_new ();
	gtk_toolbar_set_style (GTK_TOOLBAR (tbar_nav), GTK_TOOLBAR_ICONS);
	gtk_toolbar_set_orientation (GTK_TOOLBAR (tbar_nav),
				     GTK_ORIENTATION_VERTICAL);
	gtk_toolbar_set_icon_size (GTK_TOOLBAR (tbar_nav),
				   GTK_ICON_SIZE_LARGE_TOOLBAR);
	handle_nav = gtk_handle_box_new ();
	gtk_handle_box_set_handle_position (GTK_HANDLE_BOX (handle_nav),
					    GTK_POS_TOP);
	gtk_container_add (GTK_CONTAINER (handle_nav), tbar_nav);
	blast = gtk_toolbar_insert_stock (GTK_TOOLBAR (tbar_nav),
					  GTK_STOCK_GOTO_LAST,
					  _("Go to last entry"), "", move_l,
					  NULL, -1);
	bfor = gtk_toolbar_insert_stock (GTK_TOOLBAR (tbar_nav),
					 GTK_STOCK_GO_FORWARD,
					 _("Go to next entry"), "", move_f,
					 NULL, -1);
	gtk_toolbar_append_space (GTK_TOOLBAR (tbar_nav));
	bind = gtk_toolbar_insert_stock (GTK_TOOLBAR (tbar_nav),
					 GTK_STOCK_INDEX, _("View index"), "",
					 meta_list, NULL, -1);
	gtk_toolbar_append_space (GTK_TOOLBAR (tbar_nav));
	bback = gtk_toolbar_insert_stock (GTK_TOOLBAR (tbar_nav),
					  GTK_STOCK_GO_BACK,
					  _("Go to previous entry"), "",
					  move_b, NULL, -1);
	bfirst = gtk_toolbar_insert_stock (GTK_TOOLBAR (tbar_nav),
					   GTK_STOCK_GOTO_FIRST,
					   _("Go to first entry"), "", move_1,
					   NULL, -1);

	//the "main" toolbar
	tbar_main = gtk_toolbar_new ();
	gtk_toolbar_set_icon_size (GTK_TOOLBAR (tbar_main),
				   GTK_ICON_SIZE_LARGE_TOOLBAR);
	gtk_toolbar_set_style (GTK_TOOLBAR (tbar_main), GTK_TOOLBAR_ICONS);
	handle_main = gtk_handle_box_new ();
	gtk_container_add (GTK_CONTAINER (handle_main), tbar_main);

	tnew = gtk_toolbar_insert_stock (GTK_TOOLBAR (tbar_main),
					 GTK_STOCK_NEW,
					 _("New document"), "", do_new,
					 NULL, -1);

	gtk_toolbar_append_space (GTK_TOOLBAR (tbar_main));

	topen = gtk_toolbar_insert_stock (GTK_TOOLBAR (tbar_main),
					  GTK_STOCK_OPEN,
					  _("Open document"), "", meta_load,
					  NULL, -1);

	tsave = gtk_toolbar_insert_stock (GTK_TOOLBAR (tbar_main),
					  GTK_STOCK_SAVE,
					  _("Save document"), "", save,
					  NULL, -1);

	tclose = gtk_toolbar_insert_stock (GTK_TOOLBAR (tbar_main),
					   GTK_STOCK_CLOSE,
					   _("Close document"), "",
					   (GtkSignalFunc)file_close, NULL, -1);

	gtk_toolbar_append_space (GTK_TOOLBAR (tbar_main));

	tcut = gtk_toolbar_insert_stock (GTK_TOOLBAR (tbar_main),
					 GTK_STOCK_CUT,
					 _("Cut selection"), "",
					 (GtkSignalFunc) cucopa,
					 GINT_TO_POINTER (GRG_CUT), -1);

	tcopy = gtk_toolbar_insert_stock (GTK_TOOLBAR (tbar_main),
					  GTK_STOCK_COPY,
					  _("Copy selection"), "",
					  (GtkSignalFunc) cucopa,
					  GINT_TO_POINTER (GRG_COPY), -1);

	tpast = gtk_toolbar_insert_stock (GTK_TOOLBAR (tbar_main),
					  GTK_STOCK_PASTE,
					  _("Paste selection"), "",
					  (GtkSignalFunc) cucopa,
					  GINT_TO_POINTER (GRG_PASTE), -1);

	gtk_toolbar_append_space (GTK_TOOLBAR (tbar_main));

	tfind = gtk_toolbar_insert_stock (GTK_TOOLBAR (tbar_main),
					  GTK_STOCK_FIND,
					  _("Search for text"), "",
					  (GtkSignalFunc) find,
					  GINT_TO_POINTER (FALSE), -1);

	gtk_toolbar_append_space (GTK_TOOLBAR (tbar_main));

	tadd = gtk_toolbar_insert_stock (GTK_TOOLBAR (tbar_main),
					 GTK_STOCK_ADD,
					 _("Add an entry"), "", insert,
					 NULL, -1);

	trem = gtk_toolbar_insert_stock (GTK_TOOLBAR (tbar_main),
					 GTK_STOCK_REMOVE,
					 _("Remove this entry"), "", del,
					 NULL, -1);

	gtk_toolbar_append_space (GTK_TOOLBAR (tbar_main));

	tpref = gtk_toolbar_insert_stock (GTK_TOOLBAR (tbar_main),
					  GTK_STOCK_PREFERENCES,
					  _("Preferences"),
					  "", launch_prefs, NULL, -1);

	gtk_toolbar_append_space (GTK_TOOLBAR (tbar_main));

	tquit = gtk_toolbar_insert_stock (GTK_TOOLBAR (tbar_main),
					  GTK_STOCK_QUIT,
					  _("Quit Gringotts"), "",
					  meta_quit, NULL, -1);

	//by Nipo: using a GTKtable would have grown/shrunk 
	//the toolbars along with the window
	vbox = gtk_vbox_new (FALSE, 1);
	gtk_widget_show (vbox);

	menu = grg_menu_create (win1);
	gtk_box_pack_start (GTK_BOX (vbox), menu, FALSE, TRUE, 1);

	gtk_box_pack_start (GTK_BOX (vbox), handle_main, FALSE, TRUE, 1);

	hbox = gtk_hbox_new (FALSE, 1);
	gtk_box_pack_start (GTK_BOX (hbox), btitle, FALSE, TRUE, 1);
	gtk_box_pack_start (GTK_BOX (hbox), title, TRUE, TRUE, 1);
	gtk_box_pack_start (GTK_BOX (hbox), lbl, FALSE, TRUE, 1);

	gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, TRUE, 1);

	hbox = gtk_hbox_new (FALSE, 1);
	gtk_box_pack_start (GTK_BOX (hbox), handle_nav, FALSE, TRUE, 1);
	gtk_box_pack_start (GTK_BOX (hbox), scroll, TRUE, TRUE, 1);

	gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 1);

	gtk_container_add (GTK_CONTAINER (win1), vbox);
	update_saveable (GRG_SAVE_ACTIVE);
	update_saveable (GRG_SAVE_INACTIVE);
	created = TRUE;
	update ();

	//a little hack to uniform the dimension of tbar_nav and btitle
	{
		GtkRequisition gtkr1, gtkr2;
		guint bw, tw;

		gtk_widget_realize (btitle);
		gtk_widget_realize (tbar_nav);
		gtk_widget_size_request (btitle, &gtkr1);
		gtk_widget_size_request (tbar_nav, &gtkr2);
		bw = gtkr1.width;
		tw = gtkr2.width;
		if (bw > tw)
			gtk_widget_set_size_request (tbar_nav, bw, -1);
		else
			gtk_widget_set_size_request (btitle, tw, -1);
	}

	started = TRUE;
	gtk_widget_show_all (win1);
}

/**
 * main:
 * @argc: No need to tell... ;)
 * @argv: No need to tell... ;)
 *
 * No need to tell... ;)
 *
 * Returns: No need to tell... ;)
 */
gint
main (gint argc, gchar ** argv)
{
	gchar *file2load = NULL, *file2loadInArgv = NULL;
	gint prefs_err;
	gboolean root;
	
	gt = grg_initialize_defaults ("GRG", 3);
#ifdef STRONG_RANDOM
	grg_tok_set_security_lvl (gt, GRG_SEC_PARANOIA);
#endif

	//parse cmdline args
	grg_parse_argv (argc, argv, &file2loadInArgv, &root);
	
	if (!grg_security_filter (root))
	{
		grg_deinitialize (gt);
		exit (1);
	}

	grg_recent_dox_init ();
	grg_security_init ();
	prefs_err = grg_load_prefs ();

	setlocale (LC_ALL, "");
	bindtextdomain (GETTEXT_PACKAGE, LOCALEDIR);
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
	textdomain (GETTEXT_PACKAGE);

	g_print ("\n" GRG_CAP_NAME " %s " GRG_VERSION "\n\
%s\n\
\n\
%s\n\
%s\
\n", _("version"), _("(c) 2002, Germano Rizzo <mano@pluto.linux.it>"), _("released under GNU General Public License (GPL) v.2 or later"), _("See COPYING or go to http://www.gnu.org/copyleft/gpl.html"));

	gtk_init (&argc, &argv);

	clip = gtk_clipboard_get (GDK_NONE);

	grg_interface ();

	//if the preferences file is invalid, saves a default
	if (prefs_err < 0)
	{
		g_warning ("%s", _("Invalid preferences file. Resetting to defaults."));
		grg_prefs_reset_defaults ();
		grg_save_prefs ();
	}

	//loads (ev.) a startup file
	file2load = get_pref_file ();

	if(file2loadInArgv){
		file2load = g_strdup(file2loadInArgv);
		g_free(file2loadInArgv);
	}
	
	if (file2load)
	{
		if (g_file_test (file2load,
				 G_FILE_TEST_EXISTS | G_FILE_TEST_IS_REGULAR))
			load_file (file2load);
		else
			g_warning ("%s", _("File to load does not exists or it is invalid"));

		g_free (file2load);
	}

	gtk_loop_started = TRUE;
	gtk_main ();

	return (0);
}
